#include "ezdmahelper.h"

// ~2912 bytes per CircularBufferHelper
EzDmaHelper<512, 4, sizeof(void*)> helper;

// volatile
// AND only incremented in corresponding callback task
// AND only decremented in loop() task
static volatile uint32_t g_BuffersAvailableToConsume = 0;
static volatile uint32_t g_BuffersAvailableToProduce = 0;

void onBufferProducedCallback(void * pvParameters)
{
  static uint8_t readAvailableCount = 0;
  Serial.println("+ (buffer written)");
  Serial.flush();
  
  readAvailableCount++;
  __sync_fetch_and_add(&g_BuffersAvailableToConsume, 1);
}
void onBufferConsumedCallback(void * pvParameters)
{
  static uint8_t writeAvailableCount = 0;
  Serial.println("+ (buffer read)");
  Serial.flush();
  
  writeAvailableCount++;
  __sync_fetch_and_add(&g_BuffersAvailableToProduce, 1);
}

void setup(){
    Serial.begin(115200);
    while (!Serial) {
      ; // wait for serial port to connect. Needed for native USB port only
    }
    // prints title with ending line break
    Serial.println("EasyDMA Helper -- test sketch");
    Serial.flush();
    delay(1000);

    helper.begin();
    helper.onDataBufferConsumed(onBufferConsumedCallback,  nullptr);
    helper.onDataBufferProduced(onBufferProducedCallback,  nullptr);
    g_BuffersAvailableToConsume = helper.GetBufferCountReadyForConsumer();
    g_BuffersAvailableToProduce = helper.GetBufferCountReadyForProducer();
    Serial.println("Setup complete");
    Serial.flush();
    delay(1000);
}


void loop() {
    static uint32_t loopCount = 0;
    loopCount++;
    Serial.flush();
  
    // guaranteed to only get larger from callbacks...
    uint32_t buffersToConsumeCount = 0;
    while (0 != g_BuffersAvailableToConsume) {
        // interlocked/atomic decrement
        __sync_fetch_and_sub(&g_BuffersAvailableToConsume, 1);
        
        Serial.println("_ Getting buffer for CONSUMPTION");
        Serial.flush();
        
        void* buffer = helper.GetNextBufferForConsumer();

        // assert it's non-null?
        if (nullptr == buffer) {
          while(1) {
            Serial.println("_ nullptr from GetNextBufferForConsumer()?");
            Serial.flush();
            delay(5000);
          }
        }
                
        buffersToConsumeCount ++;
    }

    while (buffersToConsumeCount) {
      Serial.println("_ Indicating that CONSUME completed");
      Serial.flush();

      buffersToConsumeCount--;
      delay(200);
      helper.SetBufferConsumed();
    }

    // guaranteed to only get larger from callbacks...
    uint32_t buffersToProduceCount = 0;
    while (0 != g_BuffersAvailableToProduce) {
        // interlocked/atomic decrement
        __sync_fetch_and_sub(&g_BuffersAvailableToProduce, 1);
        Serial.println("_ Getting buffer for PRODUCTION");
        Serial.flush();

        void* buffer = helper.GetNextBufferForProducer();

        // assert it's non-null?
        if (nullptr == buffer) {
          while(1) {
            Serial.println("_ nullptr from GetNextBufferForProducer()?");
            Serial.flush();
            delay(5000);
          }
        }
                
        buffersToProduceCount ++;
    }

    while (buffersToProduceCount) {
      Serial.println("_ Indicating that PRODUCTION completed");
      Serial.flush();
      buffersToProduceCount--;
      delay(200);
      helper.SetBufferProduced();
    }   
    

}
